/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include "config.h"

#include <iostream>
#include <QApplication>
#include <QDebug>
#include <QDialogButtonBox>
#include <QLabel>
#include <QLineEdit>
#include <QPushButton>
#include <QToolButton>
#include <QUrl>
#include <QVBoxLayout>
#include "RemoteFileDialog.h"
#include "Globals.h"

#if QT_VERSION <= QT_VERSION_CHECK( 5, 14, 0 )
#define SKIP QString::SkipEmptyParts
#else
#define SKIP Qt::SkipEmptyParts
#endif

using namespace cubegui;

static QString                 defaultServer = "cube://localhost:3300";
static QHash<QString, QString> defaultDir;

QString
RemoteFileDialog::getFileName( const QString& directory, QWidget* parent )
{
    try
    {
        QString          dir = directory.isEmpty() ? defaultDir[ defaultServer ] : directory;
        RemoteFileDialog dialog( defaultServer, dir, parent );
        int              status = dialog.exec();
        dialog.disconnect();
        if ( ( status == QDialog::Accepted ) && !dialog.filename.trimmed().isEmpty() )
        {
            defaultServer = dialog.urlWidget->getUrl();
            return defaultServer + "/" + dialog.filename;
        }
    }
    catch ( const cube::NetworkError& e )
    {
        QString msg = QString( tr( "Remote file dialog: " ) ) + e.what();
        Globals::setStatusMessage( msg, Error );
    }
    return "";
}

RemoteFileDialog::RemoteFileDialog( const QString& server, const QString& directory, QWidget* parent ) : QDialog( parent )
{
    model = 0;
    proxy = new FileSortFilterProxyModel;
    tree  = new QTreeView();
    tree->setSortingEnabled( true );
    tree->sortByColumn( 0, Qt::AscendingOrder );
    tree->setModel( proxy );
    tree->setFocus();

    connect( tree, &QTreeView::activated, proxy, &FileSortFilterProxyModel::itemActivated );
    connect( tree, &QTreeView::clicked, proxy, &FileSortFilterProxyModel::itemClicked );

    this->setWindowTitle( tr( "Open file on server" ) );
    this->setModal( true );
    resize( 600, 400 );

    // server line (default: cube://localhost:3300)
    urlWidget = new UrlWidget( server );
    connect( urlWidget, &UrlWidget::urlChanged, this, &RemoteFileDialog::reconnect );

    pathWidget = new PathWidget( directory );
    connect( pathWidget, &PathWidget::pathChanged, this, &RemoteFileDialog::setModelDirectory );

    fileInput = new QLineEdit();
    connect( fileInput, &QLineEdit::returnPressed, this, &RemoteFileDialog::setFile );
    QWidget* fileLine = new QWidget;
    fileLine->setLayout( new QHBoxLayout );
    fileLine->layout()->setContentsMargins( 0, 0, 0, 0 );
    fileLine->layout()->addWidget( new QLabel( tr( "Name:" ) ) );
    fileLine->layout()->addWidget( fileInput );

    QDialogButtonBox* buttonBox = new QDialogButtonBox();
    buttonBox->addButton( QDialogButtonBox::Cancel );
    buttonBox->addButton( QDialogButtonBox::Ok );
    connect( buttonBox,  &QDialogButtonBox::rejected, this, &QDialogButtonBox::close );

    QVBoxLayout* layout = new QVBoxLayout();
    this->setContentsMargins( 0, 0, 0, 0 );
    layout->addWidget( urlWidget );
    layout->addWidget( pathWidget );
    layout->addWidget( tree );
    layout->addWidget( fileLine );
    layout->addWidget( buttonBox );

    reconnect(); // connect to cube server

    this->setLayout( layout );
    fileLine->setFocus();
}

void
RemoteFileDialog::disconnect()
{
    close();
    this->deleteLater();
    delete model;
    model = NULL;
}

void
RemoteFileDialog::fileSelected( QString fn )
{
    filename                          = fn;
    defaultDir[ urlWidget->getUrl() ] = pathWidget->getPath();
    accept();
}

/** called, if the user uses the QLineEdit "fileInput" to insert a file or directory */
void
RemoteFileDialog::setFile()
{
    QString inputPath = fileInput->text();
    if ( !inputPath.isEmpty() )
    {
        QString absolute = inputPath.startsWith( "/" ) ? inputPath : pathWidget->getPath() + "/" + inputPath;
        QString dir      = absolute.left( absolute.lastIndexOf( "/" ) );
        QString file     = absolute.mid( absolute.lastIndexOf( "/" ) + 1 );

        bool success = setModelDirectory( dir );
        if ( success )
        {
            success = setModelDirectory( absolute );
            file    = success ? "" : file;
            fileInput->setText( file ); // delete path after directory has been successfully changed
        }
        else
        {
            file = ""; // invalid path to file
            Globals::setStatusMessage( QString( tr( "Cannot read " ) ) + absolute, MessageType::Critical );
        }

        if ( !file.isEmpty() )
        {
            absolute = pathWidget->getPath() + "/" + file; // replace relative path parts after directory has been set
            if ( model->containsFile( absolute ) )
            {
                fileSelected( absolute );
            }
            else
            {
                Globals::setStatusMessage( QString( tr( "Cannot read " ) ) + absolute, MessageType::Critical );
            }
        }
    }
}

bool
RemoteFileDialog::setModelDirectory( const QString& path )
{
    bool success = false;
    if ( model )
    {
        success = model->setDirectory( path );
        if ( success )
        {
            tree->resizeColumnToContents( 0 );
            fileInput->clear();
        }
    }
    return success;
}

/** marks previous directory */
void
RemoteFileDialog::markHistory( const QString& pathStr )
{
    if ( pathStr != previousPath )
    {
        QModelIndex idx = proxy->mapFromSource( model->getModelIndex( previousPath ) );

        tree->selectionModel()->select( idx, QItemSelectionModel::Select | QItemSelectionModel::Rows );
        tree->scrollTo( idx );
        previousPath = pathStr;
    }
}

void
RemoteFileDialog::reconnect()
{
    proxy->setSourceModel( NULL ); // show empty tree
    delete model;
    model = NULL;
    try
    {
        model = new RemoteFileSystemModel( urlWidget->getUrl() );
        connect( model, &RemoteFileSystemModel::fileSelected, this, &RemoteFileDialog::fileSelected );
        connect( model, &RemoteFileSystemModel::fileClicked,  [ this ]( const QString& path ){
            QString absolute = "/" + path;
            QString file     = absolute.mid( absolute.lastIndexOf( "/" ) + 1 );
            fileInput->setText( file );
        } );
        connect( model, &RemoteFileSystemModel::directorySelected, [ this ]( const QString& path ){
            pathWidget->setPath( path );
            markHistory( path );
            fileInput->clear();
        } );

        proxy->setSourceModel( model );
        model->setDirectory( pathWidget->getPath() );
        tree->resizeColumnToContents( 0 );
    }
    catch ( const cube::NetworkError& e )
    {
        Globals::setStatusMessage( QString( tr( "Remote file dialog: " ) ) + e.what(), Error );
    }
}

//==========================================================================================

PathWidget::PathWidget( const QString& dir ) : path( dir )
{
}

/**
 * @brief PathWidget::shortenPath splits the path into a list of directories. For long pathes,
 * the directories in the middle section are united and will be hidden later
 * @param path complete path
 * @param currentPath path to the current directory
 */
QStringList
PathWidget::getPathList( const QString& path, const QString& currentPath )
{
    const int    maxwidth = 400;
    const int    minShow  = 2; // minimum count of directories to show at the beginning and the end
    QStringList  dirs;
    QFontMetrics fm( font() );
    dirs = path.split( "/", SKIP );
    bool hideParts = fm.boundingRect( path ).width() > maxwidth;
    if ( hideParts ) // for shortened paths: only show path to current directory, don't show history
    {
        dirs      = currentPath.split( "/", SKIP );
        hideParts = fm.boundingRect( currentPath ).width() > maxwidth;
    }
    int width = 0;
    if ( hideParts && dirs.size() > 2 * minShow  )
    {
        for ( int i = 0; i < minShow; i++ )
        {
            width += fm.boundingRect( dirs[ i ] ).width();
            width += fm.boundingRect( dirs[ dirs.length() - i - 1 ] ).width();
        }
        width += minShow * 2 * fm.boundingRect( " > " ).width();
        int tail = dirs.length() - minShow - 1; // first index to show after hidden elements
        while ( tail > minShow && width < maxwidth )
        {
            width += fm.boundingRect( dirs[ tail ] + " > " ).width();
            if ( width < maxwidth )
            {
                tail--;
            }
        }
        QStringList shortened;
        for ( int i = 0; i < minShow; i++ )
        {
            shortened.append( dirs[ i ] );
            width += fm.boundingRect( dirs[ i ] ).width();
        }
        QString middle; // this part will be hidden ("...")
        for ( int i = minShow; i < tail; i++ )
        {
            middle += "/" + dirs[ i ];
        }
        if ( !middle.isEmpty() )
        {
            shortened.append( middle );
        }
        for ( int i = tail; i < dirs.length(); i++ )
        {
            shortened.append( dirs[ i ] );
        }
        dirs = shortened;
    }
    return dirs;
}

void
PathWidget::setPath( const QString& pathStr )
{
    path = pathStr;
    if ( !parentSelected )
    {
        prevPath = pathStr;
    }
    parentSelected = false;

    qDeleteAll( this->findChildren<QWidget*>( QString(), Qt::FindDirectChildrenOnly ) );
    delete ( this->layout() );

    QBoxLayout* layout = new QHBoxLayout();
    setLayout( layout );
    layout->setContentsMargins( 0, 0, 0, 0 );
    layout->setSpacing( 0 );

    QFontMetrics fm( font() );
    QString      currentPath;
    QStringList  dirs = getPathList( prevPath, pathStr );

    for ( const QString& dir : dirs )
    {
        currentPath += "/" + dir;

        if ( dir.startsWith( "/" ) )
        {
            layout->addWidget( new QLabel( " > ... " ), 0, Qt::AlignLeft );
        }
        else
        {
            QPushButton* but = new QPushButton( dir );
            but->setFlat( true );
            if ( currentPath == pathStr )
            {
                but->setStyleSheet( "QPushButton{ font-weight: bold; }" );
            }
            QColor  color = palette().color( QPalette::Normal, QPalette::Highlight );
            QString col   = color.name( QColor::HexRgb );
            // ensure that flat button shows highlighted border if entered
            but->setStyleSheet( but->styleSheet() +
                                "QPushButton:hover { border: 1px solid; border-radius: 2px; border-color:" + col + "} " );

            QSize size = fm.boundingRect( but->text() ).size();
            size += QSize( 10, 10 ); // add 5 pixel margins on each side
            but->setFixedSize( size );
            connect( but, &QPushButton::clicked, [ this, currentPath ](){
                this->path = currentPath;
                parentSelected = true;
                emit pathChanged( path );
            } );
            QLabel* sep = new QLabel( ">" );
            sep->setMargin( 0 );
            layout->addWidget( sep, 0, Qt::AlignLeft );
            layout->addWidget( but, 0, Qt::AlignLeft );
        }
    }
    layout->addStretch( 1 );

    // forwards/backwards-buttons
    QToolButton* backBut = new QToolButton();
    backBut->setToolTip( tr( "Go back in history" ) );
    backBut->setIcon( QApplication::style()->standardIcon( QStyle::QStyle::SP_ArrowBack ) );
    backBut->setEnabled( !backwards.isEmpty() );
    layout->addWidget( backBut, 0, Qt::AlignRight );
    if ( !backwards.isEmpty() )
    {
        connect( backBut, &QToolButton::clicked, [ this ](){
            buttonsUsed = true;
            forwards.push( backwards.pop() ); // current dir
            this->path = backwards.pop();     // previous dir
            emit pathChanged( path );
        } );
    }

    backwards.push( pathStr );
    if ( !buttonsUsed )
    {
        forwards.clear();
    }
    buttonsUsed = false;

    QToolButton* forBut = new QToolButton();
    forBut->setToolTip( tr( "Go forward in history" ) );
    forBut->setIcon( QApplication::style()->standardIcon( QStyle::QStyle::SP_ArrowForward ) );
    forBut->setEnabled( !forwards.isEmpty() );
    layout->addWidget( forBut, 0, Qt::AlignRight );
    if ( !forwards.isEmpty() )
    {
        connect( forBut, &QToolButton::clicked, [ this ](){
            buttonsUsed = true;
            this->path = forwards.pop();
            emit pathChanged( path );
        } );
    }
}

QString
PathWidget::getPath() const
{
    return path;
}

//==========================================================================================

UrlWidget::UrlWidget( const QString& urlStr )
{
    serverLine = new QLineEdit();
    connect( serverLine, &QLineEdit::returnPressed, [ this ](){
        emit urlChanged( getUrl() );
    } );
    portLine = new QLineEdit();
    connect( portLine, &QLineEdit::returnPressed, [ this ](){
        emit urlChanged( getUrl() );
    } );
    serverLine->setFocusPolicy( Qt::ClickFocus );
    portLine->setFocusPolicy( Qt::ClickFocus );
    setUrl( urlStr );

    QWidget*     serverWidget = new QWidget();
    QHBoxLayout* hlay         = new QHBoxLayout;
    hlay->setContentsMargins( 0, 0, 0, 0 );
    serverWidget->setLayout( hlay );
    hlay->addWidget( new QLabel( tr( "Server" ) + ":" ) );
    hlay->addWidget( serverLine, 1 );
    hlay->addWidget( new QLabel( tr( "Port" ) + ":" ) );
    hlay->addWidget( portLine );
    QPushButton* but = new QPushButton( QApplication::style()->standardIcon( QStyle::SP_BrowserReload ), "" );
    but->setToolTip( tr( "Reconnect to server" ) );
    connect( but, &QPushButton::pressed, [ this ](){
        emit urlChanged( getUrl() );
    } );
    hlay->addWidget( but );

    setLayout( hlay );
}


void
UrlWidget::setUrl( const QString& urlStr )
{
    QUrl url( urlStr );
    serverLine->setText( url.host() );
    portLine->setText( QString::number( url.port() ) );
}

QString
UrlWidget::getUrl() const
{
    return "cube://" + serverLine->text().trimmed() + ":" + portLine->text().trimmed();
}
